    /*******************************************************************************
    Copyright(c) 2014 Radek Kaczorek  <rkaczorek AT gmail DOT com>
    
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License version 2 as published by the Free Software Foundation.
    .
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
    .
    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
    *******************************************************************************/
#include <stdio.h>
#include <unistd.h>
#include <memory>
#include <bcm2835.h>
#include "rpi_focus.h"

// We declare an auto pointer to focusRpi.
std::auto_ptr < FocusRpi > focusRpi (0);

// Stepper motor takes 4 miliseconds to move one step = 250 steps per second (real rate = 240,905660377)
// 1) focusing from min to max takes 7 evolutions
// 2) PG2528-0502U step motor makes 7 * (360deg/15degperstep)*72:1 = 1728 steps per evolution
// 3) MAX_STEPS for 7 evolutions should be 12096

#define MAX_STEPS 40000		// maximum steps focuser can travel from min=0 to max

#define STEP_DELAY 1		// miliseconds

// indicate GPIOs used - use P1_* pin numbers not gpio numbers




//RPi 2
#define PIN1 RPI_BPLUS_GPIO_J8_29
#define PIN2 RPI_BPLUS_GPIO_J8_31
#define PIN3 RPI_BPLUS_GPIO_J8_33
#define PIN4 RPI_BPLUS_GPIO_J8_37

static int pins[] = {
  PIN1, PIN2, PIN3, PIN4
};

namespace mode
{
  static unsigned char half[] =
  {
  8, 0b1000, 0b1100, 0b0100, 0b0110, 0b0010, 0b0011,
      0b0001, 0b1001}; 
  static unsigned char full[] =
  {
  4,0b1100, 0b0110, 0b0011, 0b1001};
};

void ISPoll (void *p);

void ISInit ()
{
  static int isInit = 0;

  if (isInit == 1)
    return;
  if (focusRpi.get () == 0)
    {
      isInit = 1;
      focusRpi.reset (new FocusRpi ());
    }
}

void ISGetProperties (const char *dev)
{
  ISInit ();
  focusRpi->ISGetProperties (dev);
}

void ISNewSwitch (const char *dev, const char *name, ISState * states,
	     char *names[], int num)
{
  ISInit ();
  focusRpi->ISNewSwitch (dev, name, states, names, num);
}

void ISNewText (const char *dev, const char *name, char *texts[], char *names[],
	   int num)
{
  ISInit ();
  focusRpi->ISNewText (dev, name, texts, names, num);
}

void ISNewNumber (const char *dev, const char *name, double values[],
	     char *names[], int num)
{
  ISInit ();
  focusRpi->ISNewNumber (dev, name, values, names, num);
}

void ISNewBLOB (const char *dev, const char *name, int sizes[], int blobsizes[],
	   char *blobs[], char *formats[], char *names[], int n)
{
  INDI_UNUSED (dev);
  INDI_UNUSED (name);
  INDI_UNUSED (sizes);
  INDI_UNUSED (blobsizes);
  INDI_UNUSED (blobs);
  INDI_UNUSED (formats);
  INDI_UNUSED (names);
  INDI_UNUSED (n);
}

void ISSnoopDevice (XMLEle * root)
{
  ISInit ();
  focusRpi->ISSnoopDevice (root);
}

FocusRpi::FocusRpi ()
{
  reverse = false;
}

FocusRpi::~FocusRpi ()
{

}

const char * FocusRpi::getDefaultName ()
{
  return (char *) "Astroberry Focuser";
}

bool FocusRpi::Connect ()
{
  int i;
  if (!bcm2835_init ())
    {
      IDMessage (getDeviceName (), "Problem initiating Astroberry Focuser.");
      return false;
    }

  // init GPIOs
  std::ofstream exportgpio;
  exportgpio.open ("/sys/class/gpio/export");
  for (i = 0; i < 4; ++i)
    {
      exportgpio << pins[i] << std::endl;
    };
  exportgpio.close ();

  // Set gpios to output mode
  for (i = 0; i < 4; ++i)
    {
      bcm2835_gpio_fsel (pins[i], BCM2835_GPIO_FSEL_OUTP);
    };

  IDMessage (getDeviceName (), "Astroberry Focuser connected successfully.");
  return true;
}

bool FocusRpi::Disconnect ()
{
  int i;
  // park focuser
  if (FocusParkingS[0].s == ISS_ON)
    {
      IDMessage (getDeviceName (), "Astroberry Focuser is parking...");
      MoveAbsFocuser (FocusAbsPosN[0].min);
    }

  // close GPIOs
  std::ofstream unexportgpio;
  unexportgpio.open ("/sys/class/gpio/unexport");
  for (i = 0; i < 4; ++i)
    {
      unexportgpio << pins[i] << std::endl;
    };
  unexportgpio.close ();
  bcm2835_close ();

  IDMessage (getDeviceName (),
	     "Astroberry Focuser disconnected successfully.");

  return true;
}

bool FocusRpi::initProperties ()
{
  INDI::Focuser::initProperties ();

  IUFillText (&PortT[0], "PORT", "Port", "RPi GPIO");
  IUFillTextVector (&PortTP, PortT, 1, getDeviceName (), "DEVICE_PORT",
		    "Ports", OPTIONS_TAB, IP_RO, 0, IPS_OK);

  IUFillNumber (&FocusAbsPosN[0], "FOCUS_ABSOLUTE_POSITION", "Ticks", "%0.0f",
		0, MAX_STEPS, (int) MAX_STEPS / 100, 0);
  IUFillNumberVector (&FocusAbsPosNP, FocusAbsPosN, 1, getDeviceName (),
		      "ABS_FOCUS_POSITION", "Position", MAIN_CONTROL_TAB,
		      IP_RW, 0, IPS_OK);

  IUFillNumber (&PresetN[0], "Preset 1", "", "%0.0f", 0, MAX_STEPS,
		(int) (MAX_STEPS / 100), 0);
  IUFillNumber (&PresetN[1], "Preset 2", "", "%0.0f", 0, MAX_STEPS,
		(int) (MAX_STEPS / 100), 0);
  IUFillNumber (&PresetN[2], "Preset 3", "", "%0.0f", 0, MAX_STEPS,
		(int) (MAX_STEPS / 100), 0);
  IUFillNumberVector (&PresetNP, PresetN, 3, getDeviceName (), "Presets",
		      "Presets", "Presets", IP_RW, 0, IPS_IDLE);

  IUFillSwitch (&PresetGotoS[0], "Preset 1", "Preset 1", ISS_OFF);
  IUFillSwitch (&PresetGotoS[1], "Preset 2", "Preset 2", ISS_OFF);
  IUFillSwitch (&PresetGotoS[2], "Preset 3", "Preset 3", ISS_OFF);
  IUFillSwitchVector (&PresetGotoSP, PresetGotoS, 3, getDeviceName (),
		      "Presets Goto", "Goto", MAIN_CONTROL_TAB, IP_RW,
		      ISR_1OFMANY, 60, IPS_OK);

  IUFillNumber (&FocusBacklashN[0], "FOCUS_BACKLASH_VALUE", "Steps", "%0.0f",
		0, (int) (MAX_STEPS / 100), (int) (MAX_STEPS / 1000), 0);
  IUFillNumberVector (&FocusBacklashNP, FocusBacklashN, 1, getDeviceName (),
		      "FOCUS_BACKLASH", "Backlash", OPTIONS_TAB, IP_RW, 0,
		      IPS_IDLE);

  IUFillSwitch (&FocusResetS[0], "FOCUS_RESET", "Reset", ISS_OFF);
  IUFillSwitchVector (&FocusResetSP, FocusResetS, 1, getDeviceName (),
		      "FOCUS_RESET", "Position Reset", OPTIONS_TAB, IP_RW,
		      ISR_1OFMANY, 60, IPS_OK);

  IUFillSwitch (&FocusParkingS[0], "FOCUS_PARKON", "Enable", ISS_OFF);
  IUFillSwitch (&FocusParkingS[1], "FOCUS_PARKOFF", "Disable", ISS_OFF);
  IUFillSwitchVector (&FocusParkingSP, FocusParkingS, 2, getDeviceName (),
		      "FOCUS_PARK", "Parking Mode", OPTIONS_TAB, IP_RW,
		      ISR_1OFMANY, 60, IPS_OK);

  // set capabilities

  FocuserCapability cap;
  cap.canAbsMove = true;
  cap.canRelMove = true;
  cap.variableSpeed = true;
  cap.canAbort = false;

  SetFocuserCapability (&cap);

  return true;
}

void FocusRpi::ISGetProperties (const char *dev)
{
  INDI::Focuser::ISGetProperties (dev);

  /* Add debug controls so we may debug driver if necessary */
  addDebugControl ();

  return;
}

bool FocusRpi::updateProperties ()
{

  INDI::Focuser::updateProperties ();

  if (isConnected ())
    {
      deleteProperty (FocusSpeedNP.name);
      defineNumber (&FocusAbsPosNP);
      defineSwitch (&FocusMotionSP);
      defineNumber (&FocusBacklashNP);
      defineSwitch (&FocusParkingSP);
      defineSwitch (&FocusResetSP);

      loadConfig ();

    }
  else
    {
      deleteProperty (FocusAbsPosNP.name);
      deleteProperty (FocusMotionSP.name);
      deleteProperty (FocusBacklashNP.name);
      deleteProperty (FocusParkingSP.name);
      deleteProperty (FocusResetSP.name);
    }

  return true;
}

bool FocusRpi::ISNewNumber (const char *dev, const char *name, double values[],
		       char *names[], int n)
{
  // first we check if it's for our device
  if (strcmp (dev, getDeviceName ()) == 0)
    {

      // handle focus absolute position
      if (!strcmp (name, FocusAbsPosNP.name))
	{
	  int newPos = (int) values[0];
	  if (MoveAbsFocuser (newPos) == IPS_OK)
	    {
	      IUUpdateNumber (&FocusAbsPosNP, values, names, n);
	      FocusAbsPosNP.s = IPS_OK;
	      IDSetNumber (&FocusAbsPosNP, NULL);
	    }
	  return true;
	}

      // handle focus relative position
      if (!strcmp (name, FocusRelPosNP.name))
	{
	  IUUpdateNumber (&FocusRelPosNP, values, names, n);

	  //FOCUS_INWARD
	  if (FocusMotionS[0].s == ISS_ON)
	    MoveRelFocuser (FOCUS_INWARD, FocusRelPosN[0].value);

	  //FOCUS_OUTWARD
	  if (FocusMotionS[1].s == ISS_ON)
	    MoveRelFocuser (FOCUS_OUTWARD, FocusRelPosN[0].value);

	  FocusRelPosNP.s = IPS_OK;
	  IDSetNumber (&FocusRelPosNP, NULL);
	  return true;
	}

      // handle focus timer
      if (!strcmp (name, FocusTimerNP.name))
	{
	  IUUpdateNumber (&FocusTimerNP, values, names, n);

	  //FOCUS_INWARD
	  if (FocusMotionS[0].s == ISS_ON)
	    MoveFocuser (FOCUS_INWARD, 0, FocusTimerN[0].value);

	  //FOCUS_OUTWARD
	  if (FocusMotionS[1].s == ISS_ON)
	    MoveFocuser (FOCUS_OUTWARD, 0, FocusTimerN[0].value);

	  FocusTimerNP.s = IPS_OK;
	  IDSetNumber (&FocusTimerNP, NULL);
	  return true;
	}

      // handle focus backlash
      if (!strcmp (name, FocusBacklashNP.name))
	{
	  IUUpdateNumber (&FocusBacklashNP, values, names, n);
	  FocusBacklashNP.s = IPS_OK;
	  IDSetNumber (&FocusBacklashNP,
		       "Astroberry Focuser backlash set to %d",
		       (int) FocusBacklashN[0].value);
	  return true;
	}
    }
  return INDI::Focuser::ISNewNumber (dev, name, values, names, n);
}

bool FocusRpi::ISNewSwitch (const char *dev, const char *name, ISState * states,
		       char *names[], int n)
{
  // first we check if it's for our device
  if (!strcmp (dev, getDeviceName ()))
    {
      /*          
         // handle focus motion in and out
         if (!strcmp(name, FocusMotionSP.name))
         {
         IUUpdateSwitch(&FocusMotionSP, states, names, n);

         //FOCUS_INWARD
         if ( FocusMotionS[0].s == ISS_ON )
         MoveRelFocuser(FOCUS_INWARD, FocusRelPosN[0].value);

         //FOCUS_OUTWARD
         if ( FocusMotionS[1].s == ISS_ON )
         MoveRelFocuser(FOCUS_OUTWARD, FocusRelPosN[0].value);

         //FocusMotionS[0].s = ISS_OFF;
         //FocusMotionS[1].s = ISS_OFF;

         FocusMotionSP.s = IPS_OK;
         IDSetSwitch(&FocusMotionSP, NULL);
         return true;
         }
       */
      // handle focus presets
      if (!strcmp (name, PresetGotoSP.name))
	{
	  IUUpdateSwitch (&PresetGotoSP, states, names, n);

	  //Preset 1
	  if (PresetGotoS[0].s == ISS_ON)
	    MoveAbsFocuser (PresetN[0].value);

	  //Preset 2
	  if (PresetGotoS[1].s == ISS_ON)
	    MoveAbsFocuser (PresetN[1].value);

	  //Preset 2
	  if (PresetGotoS[2].s == ISS_ON)
	    MoveAbsFocuser (PresetN[2].value);

	  PresetGotoS[0].s = ISS_OFF;
	  PresetGotoS[1].s = ISS_OFF;
	  PresetGotoS[2].s = ISS_OFF;
	  PresetGotoSP.s = IPS_OK;
	  IDSetSwitch (&PresetGotoSP, NULL);
	  return true;
	}

      // handle focus reset
      if (!strcmp (name, FocusResetSP.name))
	{
	  IUUpdateSwitch (&FocusResetSP, states, names, n);

	  if (FocusResetS[0].s == ISS_ON
	      && FocusAbsPosN[0].value == FocusAbsPosN[0].min)
	    {
	      FocusAbsPosN[0].value = (int) MAX_STEPS / 100;
	      IDSetNumber (&FocusAbsPosNP, NULL);
	      MoveAbsFocuser (0);
	    }
	  FocusResetS[0].s = ISS_OFF;
	  IDSetSwitch (&FocusResetSP, NULL);

	  return true;
	}

      // handle parking mode
      if (!strcmp (name, FocusParkingSP.name))
	{
	  IUUpdateSwitch (&FocusParkingSP, states, names, n);
	  IDSetSwitch (&FocusParkingSP, NULL);
	  return true;
	}

      // handle focus abort - TODO
      if (!strcmp (name, AbortSP.name))
	{
	  IUUpdateSwitch (&AbortSP, states, names, n);
	  AbortS[0].s = ISS_OFF;
	  AbortSP.s = IPS_OK;
	  IDSetSwitch (&AbortSP, NULL);
	  return true;
	}

    }
  return INDI::Focuser::ISNewSwitch (dev, name, states, names, n);
}

bool FocusRpi::ISSnoopDevice (XMLEle * root)
{
  controller->ISSnoopDevice (root);

  return INDI::Focuser::ISSnoopDevice (root);
}

bool FocusRpi::saveConfigItems (FILE * fp)
{
  IUSaveConfigText (fp, &PortTP);
  IUSaveConfigNumber (fp, &FocusRelPosNP);
  IUSaveConfigNumber (fp, &PresetNP);
  IUSaveConfigNumber (fp, &FocusBacklashNP);
  IUSaveConfigSwitch (fp, &FocusParkingSP);

  if (FocusParkingS[0].s == ISS_ON)
    IUSaveConfigNumber (fp, &FocusAbsPosNP);

  controller->saveConfigItems (fp);

  return true;
}

IPState FocusRpi::MoveFocuser (FocusDirection dir, int speed, int duration)
{
  int ticks = (int) (duration / STEP_DELAY);
  return MoveRelFocuser (dir, ticks);
}


IPState FocusRpi::MoveRelFocuser (FocusDirection dir, int ticks)
{
  int targetTicks =
    FocusAbsPosN[0].value + (ticks * (dir == FOCUS_INWARD ? -1 : 1));
  return MoveAbsFocuser (targetTicks);
}

IPState FocusRpi::MoveAbsFocuser (int targetTicks)
{
  if (targetTicks < FocusAbsPosN[0].min || targetTicks > FocusAbsPosN[0].max)
    {
      IDMessage (getDeviceName (), "Requested position is out of range.");
      return IPS_ALERT;
    }

  if (targetTicks == FocusAbsPosN[0].value)
    {
      // IDMessage(getDeviceName(), "Astroberry Focuser already in the requested position.");
      return IPS_OK;
    }

  // set focuser busy
  FocusAbsPosNP.s = IPS_BUSY;
  IDSetNumber (&FocusAbsPosNP, NULL);

  // set full step size
  SetSpeed (2);

  // check last motion direction for backlash triggering
  bool lastdir = reverse;

  // set direction
  const char *direction;
  if (targetTicks > FocusAbsPosN[0].value)
    {
      // OUTWARD
      reverse = false;
      direction = " outward ";
    }
  else
    {
      // INWARD
      reverse = true;
      direction = " inward ";
    }

  IDMessage (getDeviceName (), "Astroberry Focuser is moving %s", direction);

  // process targetTicks, if direction changed do backlash adjustment
   
  int ticks = abs (targetTicks - FocusAbsPosN[0].value+(lastdir != reverse) * FocusBacklashN[0].value ) ;

  for (int i = 0; i < ticks; i++)
    {
      stepMotor (reverse);
      FocusAbsPosN[0].value += reverse ? -1 :1;
      IDSetNumber (&FocusAbsPosNP, NULL);
    }

  // motor sleep
  disarm ();

  // update abspos value and status
  IDSetNumber (&FocusAbsPosNP, "Astroberry Focuser moved to position %0.0f",
	       FocusAbsPosN[0].value);
  FocusAbsPosNP.s = IPS_OK;
  IDSetNumber (&FocusAbsPosNP, NULL);

  return IPS_OK;
}

bool FocusRpi::SetSpeed (int speed)
{
  /* Stepper motor resolution settings (for PG2528-0502U)
   * 1) full step
   * 2) half step
   */

  switch (speed)
    {
    case 1:			// 1:1
      sequence=mode::full;
      break;
    case 2:			// 1:2
      sequence=mode::half;
      break;
    default:			// 1:1
      sequence=mode::full;
      break;
    }
  return true;
}

void FocusRpi::stepMotor (int reverse)
{
  static int i = 1;

 
  if (reverse)
    {
      i -= 1;
      if (i < 1)
	    i = sequence[0];
    }
  else
    {
      i += 1;  
    if (i > sequence[0])
	i = 1;
    }


  bcm2835_gpio_write (pins[0], sequence[i] & 0b1000);
  bcm2835_gpio_write (pins[1], sequence[i] & 0b0100);
  bcm2835_gpio_write (pins[2], sequence[i] & 0b0010);
  bcm2835_gpio_write (pins[3], sequence[i] & 0b0001);
  bcm2835_delay(STEP_DELAY);
}

void FocusRpi::disarm (void)
{
  int p = 0;
  for (p = 0; p < 4; ++p)
    {
      bcm2835_gpio_write (pins[p], 0);
    }
}
